=== modified file 'Makefile'
--- Makefile	2009-11-19 18:31:28 +0000
+++ Makefile	2010-09-09 18:28:57 +0000
@@ -209,6 +209,17 @@
 
 # Run the client with a local config and key
 run-client: all keydir/seckey.txt keydir/pubkey.txt
+	@echo "###################################################################"
+	@echo "# The following error messages are harmless and can be safely     #"
+	@echo "# ignored.  The messages are caused by not running as root, but   #"
+	@echo "# you should NOT run \"make run-client\" as root unless you also    #"
+	@echo "# unpacked and compiled Mandos as root, which is NOT recommended. #"
+	@echo "# From plugin-runner: setuid: Operation not permitted             #"
+	@echo "# From askpass-fifo:  mkfifo: Permission denied                   #"
+	@echo "# From mandos-client: setuid: Operation not permitted             #"
+	@echo "#                     seteuid: Operation not permitted            #"
+	@echo "#                     klogctl: Operation not permitted            #"
+	@echo "###################################################################"
 	./plugin-runner --plugin-dir=plugins.d \
 		--config-file=plugin-runner.conf \
 		--options-for=mandos-client:--seckey=keydir/seckey.txt,--pubkey=keydir/pubkey.txt \
@@ -221,6 +232,13 @@
 
 # Run the server with a local config
 run-server: confdir/mandos.conf confdir/clients.conf
+	@echo "#################################################################"
+	@echo "# NOTE: Please IGNORE errors about \"No permission to bind to    #"
+	@echo "# interface\" or \"Could not open file u'/var/run/mandos.pid'\" -  #"
+	@echo "# they are harmless and are caused by the server not running as #"
+	@echo "# root.  Do NOT run \"make run-server\" server as root if you did #"
+	@echo "# not also unpack and compile it as root.                       #"
+	@echo "#################################################################"
 	./mandos --debug --no-dbus --configdir=confdir $(SERVERARGS)
 
 # Used by run-server

=== modified file 'TODO'
--- TODO	2010-08-22 13:34:42 +0000
+++ TODO	2010-09-12 18:12:11 +0000
@@ -1,11 +1,13 @@
 -*- org -*-
 
-* _attribute_((nonnull))
+* Use _attribute_((nonnull)) wherever possible.
+
+* Release critical
+** Make mandos-ctl fail nicer to permission deny (runing as non-root for example)
 
 * mandos-client
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO use error() instead of perror()
 ** TODO [#B] Retry a server which has a non-definite reply:
 *** A closed connection during the TLS handshake
 *** A TCP timeout
@@ -15,33 +17,31 @@
 * splashy
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 
 * usplash
 ** TODO [#A] Make it work again
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
+** TODO Use [[info:libc:Argz%20Functions][argz_extract]]
 
 * askpass-fifo
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] Drop privileges after opening FIFO.
 
 * password-prompt
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] lock stdin (with flock()?)
 
+* plymouth
+** TODO Use [[info:libc:Argz%20Functions][argz_extract]]
+
 * TODO [#B] passdev
 
 * plugin-runner
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#C] use same file name rules as run-parts(8)
 ** kernel command line option for debug info
-** TODO [#B] use error() instead of perror()
-** TODO [$B] Use openat() and readdir64()
-   http://udrepper.livejournal.com/19395.html
+** TODO [#B] Use openat()
 
 * mandos (server)
 ** TODO [#B] Log level							  :BUGS:
@@ -75,16 +75,18 @@
 ** More D-Bus methods
 *** NeedsApproval(50, True) -> timeout, default approve
     Default approval is configurable, but True by default
-    + Approval(True) -> approve sending saved
-    + Approval(False) -> Close client connection immediately
+    + Approve(True) -> approve sending saved
+    + Approve(False) -> Close client connection immediately
 *** NeedsPassword(50) - Timeout, default disapprove
     + SetPass(u"gazonk", True) -> Approval, persistent
-    + Approval(False) -> Close client connection immediately
+    + Approve(False) -> Close client connection immediately
 ** TODO [#C] python-parsedatetime
 ** TODO [#C] systemd/launchd
    http://0pointer.de/blog/projects/systemd.html
 ** TODO Separate logging logic to own object
 ** TODO make clients to a dict!
+** TODO [#A] Limit approved_delay to max gnutls/tls timeout value
+** TODO [#B] break the wait on approved_delay if connection dies
 
 * mandos.xml
 ** [[file:mandos.xml::XXX][Document D-Bus interface]]
@@ -103,9 +105,12 @@
   arguments.
 
 * mandos-monitor
+** TODO help should be toggable
 ** Urwid client data displayer
    Better view of client data in the listing
 *** Properties popup
+** Nicer crashes. Stack traces Messes up shell.
+*** Print a nice "We are sorry" message, save stack trace to log.
 
 * mandos-keygen
 ** TODO Loop until passwords match when run interactively

=== modified file 'mandos'
--- mandos	2010-09-02 18:36:36 +0000
+++ mandos	2010-09-12 18:12:11 +0000
@@ -83,7 +83,8 @@
 
 version = "1.0.14"
 
-logger = logging.getLogger(u'mandos')
+#logger = logging.getLogger(u'mandos')
+logger = logging.Logger(u'mandos')
 syslogger = (logging.handlers.SysLogHandler
              (facility = logging.handlers.SysLogHandler.LOG_DAEMON,
               address = "/dev/log"))
@@ -98,8 +99,6 @@
                                        u' %(message)s'))
 logger.addHandler(console)
 
-multiprocessing_manager = multiprocessing.Manager()
-
 class AvahiError(Exception):
     def __init__(self, value, *args, **kwargs):
         self.value = value
@@ -158,15 +157,20 @@
                             u" after %i retries, exiting.",
                             self.rename_count)
             raise AvahiServiceError(u"Too many renames")
-        self.name = self.server.GetAlternativeServiceName(self.name)
+        self.name = unicode(self.server.GetAlternativeServiceName(self.name))
         logger.info(u"Changing Zeroconf service name to %r ...",
-                    unicode(self.name))
+                    self.name)
         syslogger.setFormatter(logging.Formatter
                                (u'Mandos (%s) [%%(process)d]:'
                                 u' %%(levelname)s: %%(message)s'
                                 % self.name))
         self.remove()
-        self.add()
+        try:
+            self.add()
+        except dbus.exceptions.DBusException, error:
+            logger.critical(u"DBusException: %s", error)
+            self.cleanup()
+            os._exit(1)
         self.rename_count += 1
     def remove(self):
         """Derived from the Avahi example code"""
@@ -195,7 +199,7 @@
         self.group.Commit()
     def entry_group_state_changed(self, state, error):
         """Derived from the Avahi example code"""
-        logger.debug(u"Avahi state change: %i", state)
+        logger.debug(u"Avahi entry group state change: %i", state)
         
         if state == avahi.ENTRY_GROUP_ESTABLISHED:
             logger.debug(u"Zeroconf service established.")
@@ -214,6 +218,7 @@
             self.group = None
     def server_state_changed(self, state):
         """Derived from the Avahi example code"""
+        logger.debug(u"Avahi server state change: %i", state)
         if state == avahi.SERVER_COLLISION:
             logger.error(u"Zeroconf server name collision")
             self.remove()
@@ -305,7 +310,6 @@
                       "rb") as secfile:
                 self.secret = secfile.read()
         else:
-            #XXX Need to allow secret on demand!
             raise TypeError(u"No secret or secfile for client %s"
                             % self.name)
         self.host = config.get(u"host", u"")
@@ -323,10 +327,10 @@
         self.checker_command = config[u"checker"]
         self.current_checker_command = None
         self.last_connect = None
-        self.approvals_pending = 0
         self._approved = None
         self.approved_by_default = config.get(u"approved_by_default",
                                               True)
+        self.approvals_pending = 0
         self.approved_delay = string_to_delta(
             config[u"approved_delay"])
         self.approved_duration = string_to_delta(
@@ -712,18 +716,19 @@
                                   + self.name.replace(u".", u"_")))
         DBusObjectWithProperties.__init__(self, self.bus,
                                           self.dbus_object_path)
-    
-    def _get_approvals_pending(value):
+
+    def _get_approvals_pending(self):
         return self._approvals_pending
-    def _set_approvals_pending(value):
+    def _set_approvals_pending(self, value):
         old_value = self._approvals_pending
         self._approvals_pending = value
         bval = bool(value)
-        if bval is bool(old_value):
+        if (hasattr(self, "dbus_object_path")
+            and bval is not bool(old_value)):
             dbus_bool = dbus.Boolean(bval, variant_level=1)
             self.PropertyChanged(dbus.String(u"approved_pending"),
                                  dbus_bool)
-        return value
+
     approvals_pending = property(_get_approvals_pending,
                                  _set_approvals_pending)
     del _get_approvals_pending, _set_approvals_pending
@@ -827,8 +832,10 @@
         return False
     
     def approve(self, value=True):
+        self.send_changedstate()
         self._approved = value
-        gobject.timeout_add(self._timedelta_to_milliseconds(self.approved_duration, self._reset_approved))
+        gobject.timeout_add(self._timedelta_to_milliseconds(self.approved_duration),
+                            self._reset_approved)
     
     
     ## D-Bus methods, signals & properties
@@ -857,7 +864,10 @@
     # GotSecret - signal
     @dbus.service.signal(_interface)
     def GotSecret(self):
-        "D-Bus signal"
+        """D-Bus signal
+        Is sent after a successful transfer of secret from the Mandos
+        server to mandos-client
+        """
         pass
     
     # Rejected - signal
@@ -1217,7 +1227,7 @@
                                            client.name)
                             if self.server.use_dbus:
                                 # Emit D-Bus signal
-                                client.Rejected("Time out")
+                                client.Rejected("Approval timed out")
                             return
                         else:
                             break
@@ -1226,8 +1236,11 @@
                 
                 sent_size = 0
                 while sent_size < len(client.secret):
-                    # XXX handle session exception
-                    sent = session.send(client.secret[sent_size:])
+                    try:
+                        sent = session.send(client.secret[sent_size:])
+                    except (gnutls.errors.GNUTLSError), error:
+                        logger.warning("gnutls send failed")
+                        return
                     logger.debug(u"Sent: %d, remaining: %d",
                                  sent, len(client.secret)
                                  - (sent_size + sent))
@@ -1243,7 +1256,10 @@
             finally:
                 if approval_required:
                     client.approvals_pending -= 1
-                session.bye()
+                try:
+                    session.bye()
+                except (gnutls.errors.GNUTLSError), error:
+                    logger.warning("gnutls bye failed")
     
     @staticmethod
     def peer_certificate(session):
@@ -1334,7 +1350,9 @@
 
         super(MultiprocessingMixInWithPipe,
               self).process_request(request, client_address)
+        self.child_pipe.close()
         self.add_pipe(parent_pipe)
+
     def add_pipe(self, parent_pipe):
         """Dummy function; override as necessary"""
         pass
@@ -1453,8 +1471,9 @@
                                        for cond, name in
                                        condition_names.iteritems()
                                        if cond & condition)
-        logger.debug(u"Handling IPC: FD = %d, condition = %s", source,
-                     conditions_string)
+        # error or the other end of multiprocessing.Pipe has closed
+        if condition & (gobject.IO_ERR | condition & gobject.IO_HUP):
+            return False
         
         # Read a request from the child
         request = parent_pipe.recv()
@@ -1503,7 +1522,7 @@
             attrname = request[1]
             value = request[2]
             setattr(client_object, attrname, value)
-        
+
         return True
 
 
@@ -1612,6 +1631,8 @@
     parser.add_option("--debug", action=u"store_true",
                       help=u"Debug mode; run in foreground and log to"
                       u" terminal")
+    parser.add_option("--debuglevel", type=u"string", metavar="Level",
+                      help=u"Debug level for stdout output")
     parser.add_option("--priority", type=u"string", help=u"GnuTLS"
                       u" priority string (see GnuTLS documentation)")
     parser.add_option("--servicename", type=u"string",
@@ -1642,6 +1663,7 @@
                         u"servicename": u"Mandos",
                         u"use_dbus": u"True",
                         u"use_ipv6": u"True",
+                        u"debuglevel": u"",
                         }
     
     # Parse config file for server-global settings
@@ -1664,7 +1686,7 @@
     # options, if set.
     for option in (u"interface", u"address", u"port", u"debug",
                    u"priority", u"servicename", u"configdir",
-                   u"use_dbus", u"use_ipv6"):
+                   u"use_dbus", u"use_ipv6", u"debuglevel"):
         value = getattr(options, option)
         if value is not None:
             server_settings[option] = value
@@ -1679,13 +1701,10 @@
     
     # For convenience
     debug = server_settings[u"debug"]
+    debuglevel = server_settings[u"debuglevel"]
     use_dbus = server_settings[u"use_dbus"]
     use_ipv6 = server_settings[u"use_ipv6"]
-    
-    if not debug:
-        syslogger.setLevel(logging.WARNING)
-        console.setLevel(logging.WARNING)
-    
+
     if server_settings[u"servicename"] != u"Mandos":
         syslogger.setFormatter(logging.Formatter
                                (u'Mandos (%s) [%%(process)d]:'
@@ -1710,7 +1729,8 @@
     tcp_server = MandosServer((server_settings[u"address"],
                                server_settings[u"port"]),
                               ClientHandler,
-                              interface=server_settings[u"interface"],
+                              interface=(server_settings[u"interface"]
+                                         or None),
                               use_ipv6=use_ipv6,
                               gnutls_priority=
                               server_settings[u"priority"],
@@ -1743,6 +1763,16 @@
             raise error
     
     # Enable all possible GnuTLS debugging
+
+
+    if not debug and not debuglevel:
+        syslogger.setLevel(logging.WARNING)
+        console.setLevel(logging.WARNING)
+    if debuglevel:
+        level = getattr(logging, debuglevel.upper())
+        syslogger.setLevel(level)
+        console.setLevel(level)
+
     if debug:
         # "Use a log level over 10 to enable all debugging options."
         # - GnuTLS manual
@@ -1754,6 +1784,16 @@
         
         (gnutls.library.functions
          .gnutls_global_set_log_function(debug_gnutls))
+
+        # Redirect stdin so all checkers get /dev/null
+        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
+        os.dup2(null, sys.stdin.fileno())
+        if null > 2:
+            os.close(null)
+    else:
+        # No console logging
+        logger.removeHandler(console)
+
     
     global main_loop
     # From the Avahi example code
@@ -1777,6 +1817,13 @@
     if server_settings["interface"]:
         service.interface = (if_nametoindex
                              (str(server_settings[u"interface"])))
+
+    if not debug:
+        # Close all input and output, do double fork, etc.
+        daemon()
+        
+    global multiprocessing_manager
+    multiprocessing_manager = multiprocessing.Manager()
     
     client_class = Client
     if use_dbus:
@@ -1800,19 +1847,7 @@
             for section in client_config.sections()))
     if not tcp_server.clients:
         logger.warning(u"No clients defined")
-    
-    if debug:
-        # Redirect stdin so all checkers get /dev/null
-        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
-        os.dup2(null, sys.stdin.fileno())
-        if null > 2:
-            os.close(null)
-    else:
-        # No console logging
-        logger.removeHandler(console)
-        # Close all input and output, do double fork, etc.
-        daemon()
-    
+        
     try:
         with pidfile:
             pid = os.getpid()

=== modified file 'mandos-ctl'
--- mandos-ctl	2009-12-26 01:19:47 +0000
+++ mandos-ctl	2010-09-12 18:12:11 +0000
@@ -8,6 +8,7 @@
 import locale
 import datetime
 import re
+import os
 
 locale.setlocale(locale.LC_ALL, u'')
 
@@ -31,11 +32,6 @@
 server_interface = domain + '.Mandos'
 client_interface = domain + '.Mandos.Client'
 version = "1.0.14"
-bus = dbus.SystemBus()
-mandos_dbus_objc = bus.get_object(busname, server_path)
-mandos_serv = dbus.Interface(mandos_dbus_objc,
-                             dbus_interface = server_interface)
-mandos_clients = mandos_serv.GetAllClientsWithProperties()
 
 def timedelta_to_milliseconds(td):
     "Convert a datetime.timedelta object to milliseconds"
@@ -93,7 +89,7 @@
         timevalue += delta
     return timevalue
 
-def print_clients(clients):
+def print_clients(clients, keywords):
     def valuetostring(value, keyword):
         if type(value) is dbus.Boolean:
             return u"Yes" if value else u"No"
@@ -113,92 +109,160 @@
     for client in clients:
         print format_string % tuple(valuetostring(client[key], key)
                                     for key in keywords)
-
-parser = OptionParser(version = "%%prog %s" % version)
-parser.add_option("-a", "--all", action="store_true",
-                  help="Print all fields")
-parser.add_option("-e", "--enable", action="store_true",
-                  help="Enable client")
-parser.add_option("-d", "--disable", action="store_true",
-                  help="disable client")
-parser.add_option("-b", "--bump-timeout", action="store_true",
-                  help="Bump timeout for client")
-parser.add_option("--start-checker", action="store_true",
-                  help="Start checker for client")
-parser.add_option("--stop-checker", action="store_true",
-                  help="Stop checker for client")
-parser.add_option("-V", "--is-enabled", action="store_true",
-                  help="Check if client is enabled")
-parser.add_option("-r", "--remove", action="store_true",
-                  help="Remove client")
-parser.add_option("-c", "--checker", type="string",
-                  help="Set checker command for client")
-parser.add_option("-t", "--timeout", type="string",
-                  help="Set timeout for client")
-parser.add_option("-i", "--interval", type="string",
-                  help="Set checker interval for client")
-parser.add_option("-H", "--host", type="string",
-                  help="Set host for client")
-parser.add_option("-s", "--secret", type="string",
-                  help="Set password blob (file) for client")
-options, client_names = parser.parse_args()
-
-# Compile list of clients to process
-clients=[]
-for name in client_names:
-    for path, client in mandos_clients.iteritems():
-        if client['name'] == name:
-            client_objc = bus.get_object(busname, path)
-            clients.append(client_objc)
-            break
-    else:
-        print >> sys.stderr, "Client not found on server: %r" % name
-        sys.exit(1)
-
-if not clients and mandos_clients.values():
-    keywords = defaultkeywords
-    if options.all:
-        keywords = ('name', 'enabled', 'timeout', 'last_checked_ok',
-                    'created', 'interval', 'host', 'fingerprint',
-                    'checker_running', 'last_enabled', 'checker')
-    print_clients(mandos_clients.values())
-
-# Process each client in the list by all selected options
-for client in clients:
-    if options.remove:
-        mandos_serv.RemoveClient(client.__dbus_object_path__)
-    if options.enable:
-        client.Enable(dbus_interface=client_interface)
-    if options.disable:
-        client.Disable(dbus_interface=client_interface)
-    if options.bump_timeout:
-        client.CheckedOK(dbus_interface=client_interface)
-    if options.start_checker:
-        client.StartChecker(dbus_interface=client_interface)
-    if options.stop_checker:
-        client.StopChecker(dbus_interface=client_interface)
-    if options.is_enabled:
-        sys.exit(0 if client.Get(client_interface,
-                                 u"enabled",
-                                 dbus_interface=dbus.PROPERTIES_IFACE)
-                 else 1)
-    if options.checker:
-        client.Set(client_interface, u"checker", options.checker,
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.host:
-        client.Set(client_interface, u"host", options.host,
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.interval:
-        client.Set(client_interface, u"interval",
-                   timedelta_to_milliseconds
-                   (string_to_delta(options.interval)),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.timeout:
-        client.Set(client_interface, u"timeout",
-                   timedelta_to_milliseconds(string_to_delta
-                                             (options.timeout)),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.secret:
-        client.Set(client_interface, u"secret",
-                   dbus.ByteArray(open(options.secret, u'rb').read()),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
+def has_actions(options):
+    return any((options.enable,
+                options.disable,
+                options.bump_timeout,
+                options.start_checker,
+                options.stop_checker,
+                options.is_enabled,
+                options.remove,
+                options.checker is not None,
+                options.timeout is not None,
+                options.interval is not None,
+                options.host is not None,
+                options.secret is not None,
+                options.approve,
+                options.deny))
+        
+def main():
+	parser = OptionParser(version = "%%prog %s" % version)
+	parser.add_option("-a", "--all", action="store_true",
+	                  help="Select all clients")
+	parser.add_option("-v", "--verbose", action="store_true",
+	                  help="Print all fields")
+	parser.add_option("-e", "--enable", action="store_true",
+	                  help="Enable client")
+	parser.add_option("-d", "--disable", action="store_true",
+	                  help="disable client")
+	parser.add_option("-b", "--bump-timeout", action="store_true",
+	                  help="Bump timeout for client")
+	parser.add_option("--start-checker", action="store_true",
+	                  help="Start checker for client")
+	parser.add_option("--stop-checker", action="store_true",
+	                  help="Stop checker for client")
+	parser.add_option("-V", "--is-enabled", action="store_true",
+	                  help="Check if client is enabled")
+	parser.add_option("-r", "--remove", action="store_true",
+	                  help="Remove client")
+	parser.add_option("-c", "--checker", type="string",
+	                  help="Set checker command for client")
+	parser.add_option("-t", "--timeout", type="string",
+	                  help="Set timeout for client")
+	parser.add_option("-i", "--interval", type="string",
+	                  help="Set checker interval for client")
+	parser.add_option("-H", "--host", type="string",
+	                  help="Set host for client")
+	parser.add_option("-s", "--secret", type="string",
+	                  help="Set password blob (file) for client")
+	parser.add_option("-A", "--approve", action="store_true",
+	                  help="Approve any current client request")
+	parser.add_option("-D", "--deny", action="store_true",
+	                  help="Deny any current client request")
+	options, client_names = parser.parse_args()
+
+        if has_actions(options) and not client_names and not options.all:
+            parser.error('Options requires clients names or --all.')
+        if options.verbose and has_actions(options):
+            parser.error('Verbose option can only be used alone or with --all.')
+        if options.all and not has_actions(options):
+            parser.error('--all requires an action')
+            
+        try:
+            bus = dbus.SystemBus()
+            mandos_dbus_objc = bus.get_object(busname, server_path)
+        except dbus.exceptions.DBusException:
+            print >> sys.stderr, "Could not connect to Mandos server"
+            sys.exit(1)
+    
+        mandos_serv = dbus.Interface(mandos_dbus_objc,
+                                     dbus_interface = server_interface)
+
+        #block stderr since dbus library prints to stderr
+        null = os.open(os.path.devnull, os.O_RDWR)
+        stderrcopy = os.dup(sys.stderr.fileno())
+        os.dup2(null, sys.stderr.fileno())
+        os.close(null)
+        try:
+            try:
+                mandos_clients = mandos_serv.GetAllClientsWithProperties()
+            finally:
+                #restore stderr
+                os.dup2(stderrcopy, sys.stderr.fileno())
+                os.close(stderrcopy)
+        except dbus.exceptions.DBusException, e:
+            print >> sys.stderr, "Access denied: Accessing mandos server through dbus."
+            sys.exit(1)
+            
+	# Compile list of clients to process
+	clients=[]
+
+        if options.all or not client_names:
+            clients = (bus.get_object(busname, path) for path in mandos_clients.iterkeys())
+        else:
+            for name in client_names:
+                for path, client in mandos_clients.iteritems():
+                    if client['name'] == name:
+                        client_objc = bus.get_object(busname, path)
+                        clients.append(client_objc)
+                        break
+                else:
+                    print >> sys.stderr, "Client not found on server: %r" % name
+                    sys.exit(1)
+            
+	if not has_actions(options) and clients:
+	    if options.verbose:
+	        keywords = ('name', 'enabled', 'timeout', 'last_checked_ok',
+	                    'created', 'interval', 'host', 'fingerprint',
+	                    'checker_running', 'last_enabled', 'checker')
+            else:
+                keywords = defaultkeywords
+                
+	    print_clients(mandos_clients.values(), keywords)
+        else:
+            # Process each client in the list by all selected options
+            for client in clients:
+                if options.remove:
+                    mandos_serv.RemoveClient(client.__dbus_object_path__)
+                if options.enable:
+                    client.Enable(dbus_interface=client_interface)
+                if options.disable:
+                    client.Disable(dbus_interface=client_interface)
+                if options.bump_timeout:
+                    client.CheckedOK(dbus_interface=client_interface)
+                if options.start_checker:
+                    client.StartChecker(dbus_interface=client_interface)
+                if options.stop_checker:
+                    client.StopChecker(dbus_interface=client_interface)
+                if options.is_enabled:
+                    sys.exit(0 if client.Get(client_interface,
+                                             u"enabled",
+                                             dbus_interface=dbus.PROPERTIES_IFACE)
+                             else 1)
+                if options.checker:
+                    client.Set(client_interface, u"checker", options.checker,
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.host:
+                    client.Set(client_interface, u"host", options.host,
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.interval:
+                    client.Set(client_interface, u"interval",
+                               timedelta_to_milliseconds
+                               (string_to_delta(options.interval)),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.timeout:
+                    client.Set(client_interface, u"timeout",
+                               timedelta_to_milliseconds(string_to_delta
+                                                         (options.timeout)),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.secret:
+                    client.Set(client_interface, u"secret",
+                               dbus.ByteArray(open(options.secret, u'rb').read()),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.approve:
+                    client.Approve(dbus.Boolean(True), dbus_interface=client_interface)
+                if options.deny:
+                    client.Approve(dbus.Boolean(False), dbus_interface=client_interface)
+
+if __name__ == '__main__':
+    main()

=== modified file 'mandos-monitor'
--- mandos-monitor	2010-08-31 18:22:25 +0000
+++ mandos-monitor	2010-09-08 19:12:04 +0000
@@ -40,6 +40,22 @@
 urwid.curses_display.curses.A_UNDERLINE |= (
     urwid.curses_display.curses.A_BLINK)
 
+def isoformat_to_datetime(iso):
+    "Parse an ISO 8601 date string to a datetime.datetime()"
+    if not iso:
+        return None
+    d, t = iso.split(u"T", 1)
+    year, month, day = d.split(u"-", 2)
+    hour, minute, second = t.split(u":", 2)
+    second, fraction = divmod(float(second), 1)
+    return datetime.datetime(int(year),
+                             int(month),
+                             int(day),
+                             int(hour),
+                             int(minute),
+                             int(second),           # Whole seconds
+                             int(fraction*1000000)) # Microseconds
+
 class MandosClientPropertyCache(object):
     """This wraps a Mandos Client D-Bus proxy object, caches the
     properties and calls a hook function when any of them are
@@ -57,8 +73,10 @@
         self.properties.update(
             self.proxy.GetAll(client_interface,
                               dbus_interface = dbus.PROPERTIES_IFACE))
-        super(MandosClientPropertyCache, self).__init__(
-            proxy_object=proxy_object, *args, **kwargs)
+
+        #XXX This break good super behaviour!
+#        super(MandosClientPropertyCache, self).__init__(
+#            *args, **kwargs)
     
     def property_changed(self, property=None, value=None):
         """This is called whenever we get a PropertyChanged signal
@@ -83,6 +101,9 @@
         # Logger
         self.logger = logger
         
+        self._update_timer_callback_tag = None
+        self.last_checker_failed = False
+        
         # The widget shown normally
         self._text_widget = urwid.Text(u"")
         # The widget shown when we have focus
@@ -112,30 +133,55 @@
                                      self.rejected,
                                      client_interface,
                                      byte_arrays=True)
+        last_checked_ok = isoformat_to_datetime(self.properties
+                                                ["last_checked_ok"])
+        if last_checked_ok is None:
+            self.last_checker_failed = True
+        else:
+            self.last_checker_failed = ((datetime.datetime.utcnow()
+                                         - last_checked_ok)
+                                        > datetime.timedelta
+                                        (milliseconds=
+                                         self.properties["interval"]))
+        if self.last_checker_failed:
+            self._update_timer_callback_tag = (gobject.timeout_add
+                                               (1000,
+                                                self.update_timer))
     
     def checker_completed(self, exitstatus, condition, command):
         if exitstatus == 0:
-            #self.logger(u'Checker for client %s (command "%s")'
-            #            u' was successful'
-            #            % (self.properties[u"name"], command))
+            if self.last_checker_failed:
+                self.last_checker_failed = False
+                gobject.source_remove(self._update_timer_callback_tag)
+                self._update_timer_callback_tag = None
+            self.logger(u'Checker for client %s (command "%s")'
+                        u' was successful'
+                        % (self.properties[u"name"], command))
+            self.update()
             return
+        # Checker failed
+        if not self.last_checker_failed:
+            self.last_checker_failed = True
+            self._update_timer_callback_tag = (gobject.timeout_add
+                                               (1000,
+                                                self.update_timer))
         if os.WIFEXITED(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' failed with exit code %s'
                         % (self.properties[u"name"], command,
                            os.WEXITSTATUS(condition)))
-            return
-        if os.WIFSIGNALED(condition):
+        elif os.WIFSIGNALED(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' was killed by signal %s'
                         % (self.properties[u"name"], command,
                            os.WTERMSIG(condition)))
-            return
-        if os.WCOREDUMP(condition):
+        elif os.WCOREDUMP(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' dumped core'
                         % (self.properties[u"name"], command))
-        self.logger(u'Checker for client %s completed mysteriously')
+        else:
+            self.logger(u'Checker for client %s completed mysteriously')
+        self.update()
     
     def checker_started(self, command):
         #self.logger(u'Client %s started checker "%s"'
@@ -143,6 +189,7 @@
         pass
     
     def got_secret(self):
+        self.last_checker_failed = False
         self.logger(u'Client %s received its secret'
                     % self.properties[u"name"])
     
@@ -184,14 +231,33 @@
                           u"bold-underline-blink":
                               u"bold-underline-blink-standout",
                           }
-        
+
         # Rebuild focus and non-focus widgets using current properties
-        self._text = (u'%(name)s: %(enabled)s'
-                      % { u"name": self.properties[u"name"],
-                          u"enabled":
-                              (u"enabled"
-                               if self.properties[u"enabled"]
-                               else u"DISABLED")})
+
+        # Base part of a client. Name!
+        base = (u'%(name)s: '
+                      % {u"name": self.properties[u"name"]})
+        if not self.properties[u"enabled"]:
+            message = u"DISABLED"
+        elif self.properties[u"approved_pending"]:
+            if self.properties[u"approved_by_default"]:
+                message = u"Connection established to client. (d)eny?"
+            else:
+                message = u"Seeks approval to send secret. (a)pprove?"
+        elif self.last_checker_failed:
+            timeout = datetime.timedelta(milliseconds
+                                         = self.properties[u"timeout"])
+            last_ok = isoformat_to_datetime(
+                max((self.properties["last_checked_ok"]
+                     or self.properties["created"]),
+                    self.properties[u"last_enabled"]))
+            timer = timeout - (datetime.datetime.utcnow() - last_ok)
+            message = (u'A checker has failed! Time until client gets diabled: %s'
+                           % unicode(timer).rsplit(".", 1)[0])
+        else:
+            message = u"enabled"
+        self._text = "%s%s" % (base, message)
+            
         if not urwid.supports_unicode():
             self._text = self._text.encode("ascii", "replace")
         textlist = [(u"normal", self._text)]
@@ -208,7 +274,15 @@
         if self.update_hook is not None:
             self.update_hook()
     
+    def update_timer(self):
+        "called by gobject"
+        self.update()
+        return True             # Keep calling this
+    
     def delete(self):
+        if self._update_timer_callback_tag is not None:
+            gobject.source_remove(self._update_timer_callback_tag)
+            self._update_timer_callback_tag = None
         if self.delete_hook is not None:
             self.delete_hook(self)
     
@@ -221,19 +295,25 @@
     def keypress(self, (maxcol,), key):
         """Handle keys.
         This overrides the method from urwid.FlowWidget"""
-        if key == u"e" or key == u"+":
-            self.proxy.Enable()
-        elif key == u"d" or key == u"-":
-            self.proxy.Disable()
+        if key == u"+":
+            self.proxy.Enable(dbus_interface = client_interface)
+        elif key == u"-":
+            self.proxy.Disable(dbus_interface = client_interface)
+        elif key == u"a":
+            self.proxy.Approve(dbus.Boolean(True, variant_level=1),
+                               dbus_interface = client_interface)
+        elif key == u"d":
+            self.proxy.Approve(dbus.Boolean(False, variant_level=1),
+                                  dbus_interface = client_interface)
         elif key == u"r" or key == u"_" or key == u"ctrl k":
             self.server_proxy_object.RemoveClient(self.proxy
                                                   .object_path)
         elif key == u"s":
-            self.proxy.StartChecker()
+            self.proxy.StartChecker(dbus_interface = client_interface)
         elif key == u"S":
-            self.proxy.StopChecker()
+            self.proxy.StopChecker(dbus_interface = client_interface)
         elif key == u"C":
-            self.proxy.CheckedOK()
+            self.proxy.CheckedOK(dbus_interface = client_interface)
         # xxx
 #         elif key == u"p" or key == "=":
 #             self.proxy.pause()
@@ -241,10 +321,10 @@
 #             self.proxy.unpause()
 #         elif key == u"RET":
 #             self.open()
-        elif key == u"+":
-            self.proxy.Approve(True)
-        elif key == u"-":
-            self.proxy.Approve(False)
+#        elif key == u"+":
+#            self.proxy.Approve(True)
+#        elif key == u"-":
+#            self.proxy.Approve(False)
         else:
             return key
     
@@ -542,12 +622,14 @@
                 self.log_message_raw((u"bold",
                                       u"  "
                                       .join((u"Clients:",
-                                             u"e: Enable",
-                                             u"d: Disable",
+                                             u"+: Enable",
+                                             u"-: Disable",
                                              u"r: Remove",
                                              u"s: Start new checker",
                                              u"S: Stop checker",
-                                             u"C: Checker OK"))))
+                                             u"C: Checker OK",
+                                             u"a: Approve",
+                                             u"d: Deny"))))
                 self.refresh()
             elif key == u"tab":
                 if self.topwidget.get_focus() is self.logbox:
@@ -581,6 +663,8 @@
 ui = UserInterface()
 try:
     ui.run()
+except KeyboardInterrupt:
+    ui.screen.stop()
 except Exception, e:
     ui.log_message(unicode(e))
     ui.screen.stop()

=== modified file 'plugin-runner.c'
--- plugin-runner.c	2010-03-27 18:39:02 +0000
+++ plugin-runner.c	2010-09-09 22:06:10 +0000
@@ -28,7 +28,7 @@
 #include <stdlib.h>		/* malloc(), exit(), EXIT_SUCCESS,
 				   realloc() */
 #include <stdbool.h>		/* bool, true, false */
-#include <stdio.h>		/* perror, fileno(), fprintf(),
+#include <stdio.h>		/* fileno(), fprintf(),
 				   stderr, STDOUT_FILENO */
 #include <sys/types.h>	        /* DIR, fdopendir(), stat(), struct
 				   stat, waitpid(), WIFEXITED(),
@@ -70,6 +70,8 @@
 #include <inttypes.h>		/* intmax_t, PRIdMAX, strtoimax() */
 #include <sysexits.h>		/* EX_OSERR, EX_USAGE, EX_IOERR,
 				   EX_CONFIG, EX_UNAVAILABLE, EX_OK */
+#include <errno.h> 		/* errno */
+#include <error.h>		/* error() */
 
 #define BUFFER_SIZE 256
 
@@ -266,7 +268,7 @@
 	/* No child processes */
 	break;
       }
-      perror("waitpid");
+      error(0, errno, "waitpid");
     }
     
     /* A child exited, find it in process_list */
@@ -357,13 +359,13 @@
   sigemptyset(&sigchld_action.sa_mask);
   ret = sigaddset(&sigchld_action.sa_mask, SIGCHLD);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     exitstatus = EX_OSERR;
     goto fallback;
   }
   ret = sigaction(SIGCHLD, &sigchld_action, &old_sigchld_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     exitstatus = EX_OSERR;
     goto fallback;
   }
@@ -599,7 +601,7 @@
   case ENOMEM:
   default:
     errno = ret;
-    perror("argp_parse");
+    error(0, errno, "argp_parse");
     exitstatus = EX_OSERR;
     goto fallback;
   case EINVAL:
@@ -626,7 +628,7 @@
     custom_argc = 1;
     custom_argv = malloc(sizeof(char*) * 2);
     if(custom_argv == NULL){
-      perror("malloc");
+      error(0, errno, "malloc");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -649,7 +651,7 @@
 	}
 	new_arg = strdup(p);
 	if(new_arg == NULL){
-	  perror("strdup");
+	  error(0, errno, "strdup");
 	  exitstatus = EX_OSERR;
 	  free(org_line);
 	  goto fallback;
@@ -659,7 +661,7 @@
 	custom_argv = realloc(custom_argv, sizeof(char *)
 			      * ((unsigned int) custom_argc + 1));
 	if(custom_argv == NULL){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  exitstatus = EX_OSERR;
 	  free(org_line);
 	  goto fallback;
@@ -672,7 +674,7 @@
       ret = fclose(conffp);
     } while(ret == EOF and errno == EINTR);
     if(ret == EOF){
-      perror("fclose");
+      error(0, errno, "fclose");
       exitstatus = EX_IOERR;
       goto fallback;
     }
@@ -681,7 +683,7 @@
     /* Check for harmful errors and go to fallback. Other errors might
        not affect opening plugins */
     if(errno == EMFILE or errno == ENFILE or errno == ENOMEM){
-      perror("fopen");
+      error(0, errno, "fopen");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -698,7 +700,7 @@
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      error(0, errno, "argp_parse");
       exitstatus = EX_OSERR;
       goto fallback;
     case EINVAL:
@@ -718,7 +720,7 @@
   case ENOMEM:
   default:
     errno = ret;
-    perror("argp_parse");
+    error(0, errno, "argp_parse");
     exitstatus = EX_OSERR;
     goto fallback;
   case EINVAL:
@@ -743,11 +745,11 @@
   /* Strip permissions down to nobody */
   setgid(gid);
   if(ret == -1){
-    perror("setgid");
+    error(0, errno, "setgid");
   }
   ret = setuid(uid);
   if(ret == -1){
-    perror("setuid");
+    error(0, errno, "setuid");
   }
   
   /* Open plugin directory with close_on_exec flag */
@@ -771,7 +773,7 @@
 		    );
     }
     if(dir_fd == -1){
-      perror("Could not open plugin dir");
+      error(0, errno, "Could not open plugin dir");
       exitstatus = EX_UNAVAILABLE;
       goto fallback;
     }
@@ -780,7 +782,7 @@
   /* Set the FD_CLOEXEC flag on the directory */
     ret = set_cloexec_flag(dir_fd);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
@@ -789,7 +791,7 @@
     
     dir = fdopendir(dir_fd);
     if(dir == NULL){
-      perror("Could not open plugin dir");
+      error(0, errno, "Could not open plugin dir");
       TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
@@ -807,7 +809,7 @@
     /* All directory entries have been processed */
     if(dirst == NULL){
       if(errno == EBADF){
-	perror("readdir");
+	error(0, errno, "readdir");
 	exitstatus = EX_IOERR;
 	goto fallback;
       }
@@ -870,13 +872,13 @@
 					     dirst->d_name));
     }
     if(ret < 0){
-      perror("asprintf");
+      error(0, errno, "asprintf");
       continue;
     }
     
     ret = (int)TEMP_FAILURE_RETRY(stat(filename, &st));
     if(ret == -1){
-      perror("stat");
+      error(0, errno, "stat");
       free(filename);
       continue;
     }
@@ -894,7 +896,7 @@
     
     plugin *p = getplugin(dirst->d_name);
     if(p == NULL){
-      perror("getplugin");
+      error(0, errno, "getplugin");
       free(filename);
       continue;
     }
@@ -912,13 +914,13 @@
       if(g != NULL){
 	for(char **a = g->argv + 1; *a != NULL; a++){
 	  if(not add_argument(p, *a)){
-	    perror("add_argument");
+	    error(0, errno, "add_argument");
 	  }
 	}
 	/* Add global environment variables */
 	for(char **e = g->environ; *e != NULL; e++){
 	  if(not add_environment(p, *e, false)){
-	    perror("add_environment");
+	    error(0, errno, "add_environment");
 	  }
 	}
       }
@@ -929,7 +931,7 @@
     if(p->environ[0] != NULL){
       for(char **e = environ; *e != NULL; e++){
 	if(not add_environment(p, *e, false)){
-	  perror("add_environment");
+	  error(0, errno, "add_environment");
 	}
       }
     }
@@ -937,20 +939,20 @@
     int pipefd[2];
     ret = (int)TEMP_FAILURE_RETRY(pipe(pipefd));
     if(ret == -1){
-      perror("pipe");
+      error(0, errno, "pipe");
       exitstatus = EX_OSERR;
       goto fallback;
     }
     /* Ask OS to automatic close the pipe on exec */
     ret = set_cloexec_flag(pipefd[0]);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       exitstatus = EX_OSERR;
       goto fallback;
     }
     ret = set_cloexec_flag(pipefd[1]);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -959,7 +961,7 @@
 					      &sigchld_action.sa_mask,
 					      NULL));
     if(ret < 0){
-      perror("sigprocmask");
+      error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -969,7 +971,7 @@
       pid = fork();
     } while(pid == -1 and errno == EINTR);
     if(pid == -1){
-      perror("fork");
+      error(0, errno, "fork");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -977,18 +979,18 @@
       /* this is the child process */
       ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
       if(ret < 0){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	_exit(EX_OSERR);
       }
       ret = sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
       if(ret < 0){
-	perror("sigprocmask");
+	error(0, errno, "sigprocmask");
 	_exit(EX_OSERR);
       }
       
       ret = dup2(pipefd[1], STDOUT_FILENO); /* replace our stdout */
       if(ret == -1){
-	perror("dup2");
+	error(0, errno, "dup2");
 	_exit(EX_OSERR);
       }
       
@@ -999,12 +1001,12 @@
       }
       if(p->environ[0] == NULL){
 	if(execv(filename, p->argv) < 0){
-	  perror("execv");
+	  error(0, errno, "execv for %s", filename);
 	  _exit(EX_OSERR);
 	}
       } else {
 	if(execve(filename, p->argv, p->environ) < 0){
-	  perror("execve");
+	  error(0, errno, "execve for %s", filename);
 	  _exit(EX_OSERR);
 	}
       }
@@ -1016,12 +1018,12 @@
     free(filename);
     plugin *new_plugin = getplugin(dirst->d_name);
     if(new_plugin == NULL){
-      perror("getplugin");
+      error(0, errno, "getplugin");
       ret = (int)(TEMP_FAILURE_RETRY
 		  (sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask,
 			       NULL)));
       if(ret < 0){
-        perror("sigprocmask");
+        error(0, errno, "sigprocmask");
       }
       exitstatus = EX_OSERR;
       goto fallback;
@@ -1036,7 +1038,7 @@
 					      &sigchld_action.sa_mask,
 					      NULL));
     if(ret < 0){
-      perror("sigprocmask");
+      error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1069,7 +1071,7 @@
     fd_set rfds = rfds_all;
     int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
     if(select_ret == -1 and errno != EINTR){
-      perror("select");
+      error(0, errno, "select");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1111,7 +1113,7 @@
 					 &sigchld_action.sa_mask,
 					 NULL));
 	  if(ret < 0){
-	    perror("sigprocmask");
+	    error(0, errno, "sigprocmask");
 	    exitstatus = EX_OSERR;
 	    goto fallback;
 	  }
@@ -1125,7 +1127,7 @@
 		      (sigprocmask(SIG_UNBLOCK,
 				   &sigchld_action.sa_mask, NULL)));
 	  if(ret < 0){
-	    perror("sigprocmask");
+	    error(0, errno, "sigprocmask");
 	    exitstatus = EX_OSERR;
 	    goto fallback;
 	  }
@@ -1142,7 +1144,7 @@
 	bool bret = print_out_password(proc->buffer,
 				       proc->buffer_length);
 	if(not bret){
-	  perror("print_out_password");
+	  error(0, errno, "print_out_password");
 	  exitstatus = EX_IOERR;
 	}
 	goto fallback;
@@ -1161,7 +1163,7 @@
 	proc->buffer = realloc(proc->buffer, proc->buffer_size
 			       + (size_t) BUFFER_SIZE);
 	if(proc->buffer == NULL){
-	  perror("malloc");
+	  error(0, errno, "malloc");
 	  exitstatus = EX_OSERR;
 	  goto fallback;
 	}
@@ -1204,7 +1206,7 @@
     }
     bret = print_out_password(passwordbuffer, len);
     if(not bret){
-      perror("print_out_password");
+      error(0, errno, "print_out_password");
       exitstatus = EX_IOERR;
     }
   }
@@ -1212,7 +1214,7 @@
   /* Restore old signal handler */
   ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     exitstatus = EX_OSERR;
   }
   
@@ -1234,7 +1236,7 @@
       ret = kill(p->pid, SIGTERM);
       if(ret == -1 and errno != ESRCH){
 	/* Set-uid proccesses might not get closed */
-	perror("kill");
+	error(0, errno, "kill");
       }
     }
   }
@@ -1244,7 +1246,7 @@
     ret = wait(NULL);
   } while(ret >= 0);
   if(errno != ECHILD){
-    perror("wait");
+    error(0, errno, "wait");
   }
   
   free_plugin_list();

=== modified file 'plugins.d/askpass-fifo.c'
--- plugins.d/askpass-fifo.c	2009-10-18 16:09:05 +0000
+++ plugins.d/askpass-fifo.c	2010-09-07 16:48:58 +0000
@@ -31,7 +31,7 @@
 				   ENOENT, EEXIST, EFAULT, EMFILE,
 				   ENFILE, ENOMEM, EBADF, EINVAL, EIO,
 				   EISDIR, EFBIG */
-#include <stdio.h>		/* perror() */
+#include <error.h>		/* error() */
 #include <stdlib.h>		/* EXIT_FAILURE, NULL, size_t, free(),
 				   realloc(), EXIT_SUCCESS */
 #include <fcntl.h>		/* open(), O_RDONLY */
@@ -51,7 +51,7 @@
   ret = mkfifo(passfifo, S_IRUSR | S_IWUSR);
   if(ret == -1){
     int e = errno;
-    perror("mkfifo");
+    error(0, errno, "mkfifo");
     switch(e){
     case EACCES:
     case ENOTDIR:
@@ -73,7 +73,7 @@
   int fifo_fd = open(passfifo, O_RDONLY);
   if(fifo_fd == -1){
     int e = errno;
-    perror("open");
+    error(0, errno, "open");
     switch(e){
     case EACCES:
     case ENOENT:
@@ -101,7 +101,7 @@
       if(buf_len + blocksize > buf_allocated){
 	char *tmp = realloc(buf, buf_allocated + blocksize);
 	if(tmp == NULL){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  free(buf);
 	  return EX_OSERR;
 	}
@@ -113,7 +113,7 @@
 	int e = errno;
 	free(buf);
 	errno = e;
-	perror("read");
+	error(0, errno, "read");
 	switch(e){
 	case EBADF:
 	case EFAULT:
@@ -141,7 +141,7 @@
       int e = errno;
       free(buf);
       errno = e;
-      perror("write");
+      error(0, errno, "write");
       switch(e){
       case EBADF:
       case EFAULT:
@@ -161,7 +161,7 @@
   ret = close(STDOUT_FILENO);
   if(ret == -1){
     int e = errno;
-    perror("close");
+    error(0, errno, "close");
     switch(e){
     case EBADF:
       return EX_OSFILE;

=== modified file 'plugins.d/mandos-client.c'
--- plugins.d/mandos-client.c	2009-11-01 00:10:28 +0000
+++ plugins.d/mandos-client.c	2010-09-12 18:12:11 +0000
@@ -63,7 +63,7 @@
 				   strtoimax() */
 #include <assert.h>		/* assert() */
 #include <errno.h>		/* perror(), errno */
-#include <time.h>		/* nanosleep(), time() */
+#include <time.h>		/* nanosleep(), time(), sleep() */
 #include <net/if.h>		/* ioctl, ifreq, SIOCGIFFLAGS, IFF_UP,
 				   SIOCSIFFLAGS, if_indextoname(),
 				   if_nametoindex(), IF_NAMESIZE */
@@ -127,6 +127,8 @@
 static const char mandos_protocol_version[] = "1";
 const char *argp_program_version = "mandos-client " VERSION;
 const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
+static const char sys_class_net[] = "/sys/class/net";
+char *connect_to = NULL;
 
 /* Used for passing in values through the Avahi callback functions */
 typedef struct {
@@ -688,9 +690,11 @@
     ret = connect(tcp_sd, &to.in, sizeof(to)); /* IPv4 */
   }
   if(ret < 0){
-    int e = errno;
-    perror("connect");
-    errno = e;
+    if ((errno != ECONNREFUSED and errno != ENETUNREACH) or debug){
+      int e = errno;
+      perror("connect");
+      errno = e;
+    }
     goto mandos_end;
   }
   
@@ -1020,6 +1024,101 @@
   errno = old_errno;
 }
 
+/* 
+ * This function determines if a directory entry in /sys/class/net
+ * corresponds to an acceptable network device.
+ * (This function is passed to scandir(3) as a filter function.)
+ */
+int good_interface(const struct dirent *if_entry){
+  ssize_t ssret;
+  char *flagname = NULL;
+  int ret = asprintf(&flagname, "%s/%s/flags", sys_class_net,
+		     if_entry->d_name);
+  if(ret < 0){
+    perror("asprintf");
+    return 0;
+  }
+  if(if_entry->d_name[0] == '.'){
+    return 0;
+  }
+  int flags_fd = (int)TEMP_FAILURE_RETRY(open(flagname, O_RDONLY));
+  if(flags_fd == -1){
+    perror("open");
+    return 0;
+  }
+  typedef short ifreq_flags;	/* ifreq.ifr_flags in netdevice(7) */
+  /* read line from flags_fd */
+  ssize_t to_read = (sizeof(ifreq_flags)*2)+3; /* "0x1003\n" */
+  char *flagstring = malloc((size_t)to_read+1); /* +1 for final \0 */
+  flagstring[(size_t)to_read] = '\0';
+  if(flagstring == NULL){
+    perror("malloc");
+    close(flags_fd);
+    return 0;
+  }
+  while(to_read > 0){
+    ssret = (ssize_t)TEMP_FAILURE_RETRY(read(flags_fd, flagstring,
+					     (size_t)to_read));
+    if(ssret == -1){
+      perror("read");
+      free(flagstring);
+      close(flags_fd);
+      return 0;
+    }
+    to_read -= ssret;
+    if(ssret == 0){
+      break;
+    }
+  }
+  close(flags_fd);
+  intmax_t tmpmax;
+  char *tmp;
+  errno = 0;
+  tmpmax = strtoimax(flagstring, &tmp, 0);
+  if(errno != 0 or tmp == flagstring or (*tmp != '\0'
+					 and not (isspace(*tmp)))
+     or tmpmax != (ifreq_flags)tmpmax){
+    if(debug){
+      fprintf(stderr, "Invalid flags \"%s\" for interface \"%s\"\n",
+	      flagstring, if_entry->d_name);
+    }
+    free(flagstring);
+    return 0;
+  }
+  free(flagstring);
+  ifreq_flags flags = (ifreq_flags)tmpmax;
+  /* Reject the loopback device */
+  if(flags & IFF_LOOPBACK){
+    if(debug){
+      fprintf(stderr, "Rejecting loopback interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  /* Accept point-to-point devices only if connect_to is specified */
+  if(connect_to != NULL and (flags & IFF_POINTOPOINT)){
+    if(debug){
+      fprintf(stderr, "Accepting point-to-point interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 1;
+  }
+  /* Otherwise, reject non-broadcast-capable devices */
+  if(not (flags & IFF_BROADCAST)){
+    if(debug){
+      fprintf(stderr, "Rejecting non-broadcast interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  /* Accept this device */
+  if(debug){
+    fprintf(stderr, "Interface \"%s\" is acceptable\n",
+	    if_entry->d_name);
+  }
+  return 1;
+}
+
 int main(int argc, char *argv[]){
   AvahiSServiceBrowser *sb = NULL;
   int error;
@@ -1027,13 +1126,12 @@
   intmax_t tmpmax;
   char *tmp;
   int exitcode = EXIT_SUCCESS;
-  const char *interface = "eth0";
+  const char *interface = "";
   struct ifreq network;
   int sd = -1;
   bool take_down_interface = false;
   uid_t uid;
   gid_t gid;
-  char *connect_to = NULL;
   char tempdir[] = "/tmp/mandosXXXXXX";
   bool tempdir_created = false;
   AvahiIfIndex if_index = AVAHI_IF_UNSPEC;
@@ -1196,6 +1294,31 @@
   if(not debug){
     avahi_set_log_function(empty_log);
   }
+
+  if(interface[0] == '\0'){
+    struct dirent **direntries;
+    ret = scandir(sys_class_net, &direntries, good_interface,
+		  alphasort);
+    if(ret >= 1){
+      /* Pick the first good interface */
+      interface = strdup(direntries[0]->d_name);
+      if(debug){
+	fprintf(stderr, "Using interface \"%s\"\n", interface);
+      }
+      if(interface == NULL){
+	perror("malloc");
+	free(direntries);
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      free(direntries);
+    } else {
+      free(direntries);
+      fprintf(stderr, "Could not find a network interface\n");
+      exitcode = EXIT_FAILURE;
+      goto end;
+    }
+  }
   
   /* Initialize Avahi early so avahi_simple_poll_quit() can be called
      from the signal handler */
@@ -1272,7 +1395,7 @@
   }
   
   /* If the interface is down, bring it up */
-  if(interface[0] != '\0'){
+  if(strcmp(interface, "none") != 0){
     if_index = (AvahiIfIndex) if_nametoindex(interface);
     if(if_index == 0){
       fprintf(stderr, "No such interface: \"%s\"\n", interface);
@@ -1349,7 +1472,7 @@
       ret = ioctl(sd, SIOCSIFFLAGS, &network);
       if(ret == -1){
 	take_down_interface = false;
-	perror("ioctl SIOCSIFFLAGS");
+	perror("ioctl SIOCSIFFLAGS +IFF_UP");
 	exitcode = EX_OSERR;
 #ifdef __linux__
 	if(restore_loglevel){
@@ -1497,31 +1620,19 @@
     if(quit_now){
       goto end;
     }
-    
-    ret = start_mandos_communication(address, port, if_index, af);
-    if(ret < 0){
-      switch(errno){
-      case ENETUNREACH:
-      case EHOSTDOWN:
-      case EHOSTUNREACH:
-	exitcode = EX_NOHOST;
-	break;
-      case EINVAL:
-	exitcode = EX_USAGE;
-	break;
-      case EIO:
-	exitcode = EX_IOERR;
-	break;
-      case EPROTO:
-	exitcode = EX_PROTOCOL;
-	break;
-      default:
-	exitcode = EX_OSERR;
+
+    while(not quit_now){
+      ret = start_mandos_communication(address, port, if_index, af);
+      if(quit_now or ret == 0){
 	break;
       }
-    } else {
+      sleep(15);
+    };
+
+    if (not quit_now){
       exitcode = EXIT_SUCCESS;
     }
+
     goto end;
   }
   
@@ -1624,7 +1735,7 @@
 	network.ifr_flags &= ~(short)IFF_UP; /* clear flag */
 	ret = ioctl(sd, SIOCSIFFLAGS, &network);
 	if(ret == -1){
-	  perror("ioctl SIOCSIFFLAGS");
+	  perror("ioctl SIOCSIFFLAGS -IFF_UP");
 	}
       }
       ret = (int)TEMP_FAILURE_RETRY(close(sd));

=== modified file 'plugins.d/mandos-client.xml'
--- plugins.d/mandos-client.xml	2009-02-09 02:01:13 +0000
+++ plugins.d/mandos-client.xml	2010-08-10 19:08:24 +0000
@@ -202,8 +202,9 @@
 	<listitem>
 	  <para>
 	    Network interface that will be brought up and scanned for
-	    Mandos servers to connect to.  The default is
-	    <quote><literal>eth0</literal></quote>.
+	    Mandos servers to connect to.  The default is the empty
+	    string, which will automatically choose an appropriate
+	    interface.
 	  </para>
 	  <para>
 	    If the <option>--connect</option> option is used, this
@@ -220,10 +221,11 @@
 	    by this program.
 	  </para>
 	  <para>
-	    <replaceable>NAME</replaceable> can be the empty string;
-	    this will not use any specific interface, and will not
-	    bring up an interface on startup.  This is not
-	    recommended, and only meant for advanced users.
+	    <replaceable>NAME</replaceable> can be the string
+	    <quote><literal>none</literal></quote>; this will not use
+	    any specific interface, and will not bring up an interface
+	    on startup.  This is not recommended, and only meant for
+	    advanced users.
 	  </para>
 	</listitem>
       </varlistentry>

=== modified file 'plugins.d/password-prompt.c'
--- plugins.d/password-prompt.c	2010-03-27 18:39:02 +0000
+++ plugins.d/password-prompt.c	2010-09-07 16:48:58 +0000
@@ -39,11 +39,12 @@
 #include <stdlib.h>		/* EXIT_SUCCESS, EXIT_FAILURE,
 				   getenv() */
 #include <stdio.h>		/* fprintf(), stderr, getline(),
-				   stdin, feof(), perror(), fputc()
+				   stdin, feof(), fputc()
 				*/
 #include <errno.h>		/* errno, EBADF, ENOTTY, EINVAL,
 				   EFAULT, EFBIG, EIO, ENOSPC, EINTR
 				*/
+#include <error.h>		/* error() */
 #include <iso646.h>		/* or, not */
 #include <stdbool.h>		/* bool, false, true */
 #include <string.h> 		/* strlen, rindex */
@@ -70,7 +71,8 @@
 }
 
 int main(int argc, char **argv){
-  ssize_t ret;
+  ssize_t sret;
+  int ret;
   size_t n;
   struct termios t_new, t_old;
   char *buffer = NULL;
@@ -139,7 +141,7 @@
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      error(0, errno, "argp_parse");
       return EX_OSERR;
     case EINVAL:
       return EX_USAGE;
@@ -155,7 +157,7 @@
   
   if(tcgetattr(STDIN_FILENO, &t_old) != 0){
     int e = errno;
-    perror("tcgetattr");
+    error(0, errno, "tcgetattr");
     switch(e){
     case EBADF:
     case ENOTTY:
@@ -168,17 +170,17 @@
   sigemptyset(&new_action.sa_mask);
   ret = sigaddset(&new_action.sa_mask, SIGINT);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   ret = sigaddset(&new_action.sa_mask, SIGHUP);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   ret = sigaddset(&new_action.sa_mask, SIGTERM);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   /* Need to check if the handler is SIG_IGN before handling:
@@ -187,37 +189,37 @@
   */
   ret = sigaction(SIGINT, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGINT, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
   ret = sigaction(SIGHUP, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGHUP, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
   ret = sigaction(SIGTERM, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGTERM, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
@@ -231,7 +233,7 @@
   t_new.c_lflag &= ~(tcflag_t)ECHO;
   if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_new) != 0){
     int e = errno;
-    perror("tcsetattr-echo");
+    error(0, errno, "tcsetattr-echo");
     switch(e){
     case EBADF:
     case ENOTTY:
@@ -286,11 +288,11 @@
 	}
       }
     }
-    ret = getline(&buffer, &n, stdin);
-    if(ret > 0){
+    sret = getline(&buffer, &n, stdin);
+    if(sret > 0){
       status = EXIT_SUCCESS;
       /* Make n = data size instead of allocated buffer size */
-      n = (size_t)ret;
+      n = (size_t)sret;
       /* Strip final newline */
       if(n > 0 and buffer[n-1] == '\n'){
 	buffer[n-1] = '\0';	/* not strictly necessary */
@@ -298,10 +300,10 @@
       }
       size_t written = 0;
       while(written < n){
-	ret = write(STDOUT_FILENO, buffer + written, n - written);
-	if(ret < 0){
+	sret = write(STDOUT_FILENO, buffer + written, n - written);
+	if(sret < 0){
 	  int e = errno;
-	  perror("write");
+	  error(0, errno, "write");
 	  switch(e){
 	  case EBADF:
 	  case EFAULT:
@@ -318,12 +320,12 @@
 	  }
 	  break;
 	}
-	written += (size_t)ret;
+	written += (size_t)sret;
       }
-      ret = close(STDOUT_FILENO);
-      if(ret == -1){
+      sret = close(STDOUT_FILENO);
+      if(sret == -1){
 	int e = errno;
-	perror("close");
+	error(0, errno, "close");
 	switch(e){
 	case EBADF:
 	  status = EX_OSFILE;
@@ -336,10 +338,10 @@
       }
       break;
     }
-    if(ret < 0){
+    if(sret < 0){
       int e = errno;
       if(errno != EINTR and not feof(stdin)){
-	perror("getline");
+	error(0, errno, "getline");
 	switch(e){
 	case EBADF:
 	  status = EX_UNAVAILABLE;
@@ -352,7 +354,7 @@
 	break;
       }
     }
-    /* if(ret == 0), then the only sensible thing to do is to retry to
+    /* if(sret == 0), then the only sensible thing to do is to retry to
        read from stdin */
     fputc('\n', stderr);
     if(debug and not quit_now){
@@ -368,7 +370,7 @@
     fprintf(stderr, "Restoring terminal attributes\n");
   }
   if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_old) != 0){
-    perror("tcsetattr+echo");
+    error(0, errno, "tcsetattr+echo");
   }
   
   if(quit_now){
@@ -376,7 +378,7 @@
     old_action.sa_handler = SIG_DFL;
     ret = sigaction(signal_received, &old_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     raise(signal_received);
   }

=== added file 'plugins.d/plymouth.c'
--- plugins.d/plymouth.c	1970-01-01 00:00:00 +0000
+++ plugins.d/plymouth.c	2010-09-07 18:48:56 +0000
@@ -0,0 +1,405 @@
+#define _GNU_SOURCE		/* asprintf(), TEMP_FAILURE_RETRY() */
+#include <signal.h>		/* sig_atomic_t, struct sigaction,
+				   sigemptyset(), sigaddset(), SIGINT,
+				   SIGHUP, SIGTERM, sigaction(),
+				   kill(), SIG_IGN */
+#include <stdbool.h>		/* bool, false, true */
+#include <fcntl.h>		/* open(), O_RDONLY */
+#include <iso646.h>		/* and, or, not*/
+#include <sys/types.h>		/* size_t, ssize_t, pid_t, struct dirent,
+				   waitpid() */
+#include <sys/wait.h>		/* waitpid() */
+#include <stddef.h>		/* NULL */
+#include <string.h>		/* strchr(), memcmp() */
+#include <stdio.h>		/* asprintf(), perror(), fopen(), fscanf() */
+#include <unistd.h>		/* close(), readlink(), read(), fork()
+				   setsid(), chdir(), dup2()
+				   STDERR_FILENO, execv(), access() */
+#include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
+#include <dirent.h>		/* scandir(), alphasort() */
+#include <inttypes.h>		/* intmax_t, strtoumax(), SCNuMAX */
+#include <sys/stat.h>		/* struct stat, lstat() */
+#include <sysexits.h>		/* EX_OSERR */
+#include <error.h>		/* error() */
+#include <errno.h>		/* TEMP_FAILURE_RETRY */
+#include <stdarg.h>
+
+sig_atomic_t interrupted_by_signal = 0;
+const char plymouth_pid[] = "/dev/.initramfs/plymouth.pid";
+const char plymouth_path[] = "/bin/plymouth";
+const char plymouthd_path[] = "/sbin/plymouthd";
+const char *plymouthd_default_argv[] = {"/sbin/plymouthd", "--mode=boot",
+					"--attach-to-session",
+					"--pid-file=/dev/.initramfs/plymouth.pid",
+					NULL };
+
+static void termination_handler(__attribute__((unused))int signum){
+  if(interrupted_by_signal){
+    return;
+  }
+  interrupted_by_signal = 1;
+}
+
+/* Create prompt string */
+char *makeprompt(void){
+  int ret = 0;
+  char *prompt;
+  const char *const cryptsource = getenv("cryptsource");
+  const char *const crypttarget = getenv("crypttarget");
+  const char prompt_start[] = "Enter passphrase to unlock the disk";
+  
+  if(cryptsource == NULL){
+    if(crypttarget == NULL){
+      ret = asprintf(&prompt, "%s: ", prompt_start);
+    } else {
+      ret = asprintf(&prompt, "%s (%s): ", prompt_start,
+		     crypttarget);
+    }
+  } else {
+    if(crypttarget == NULL){
+      ret = asprintf(&prompt, "%s %s: ", prompt_start, cryptsource);
+    } else {
+      ret = asprintf(&prompt, "%s %s (%s): ", prompt_start,
+		     cryptsource, crypttarget);
+    }
+  }
+  if(ret == -1){
+    return NULL;
+  }
+  return prompt;
+}
+
+void kill_and_wait(pid_t pid){
+  TEMP_FAILURE_RETRY(kill(pid, SIGTERM));
+  TEMP_FAILURE_RETRY(waitpid(pid, NULL, 0));
+}
+
+bool become_a_daemon(void){
+  int ret = setuid(geteuid());
+  if(ret == -1){
+    error(0, errno, "setuid");
+  }
+    
+  setsid();
+  ret = chdir("/");
+  if(ret == -1){
+    error(0, errno, "chdir");
+    return false;
+  }
+  ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
+  if(ret == -1){
+    error(0, errno, "dup2");
+    return false;
+  }
+  return true;
+}
+
+bool exec_and_wait(pid_t *pid_return, const char *path,
+		   const char **argv, bool interruptable,
+		   bool daemonize){
+  int status;
+  int ret;
+  pid_t pid;
+  pid = fork();
+  if(pid == -1){
+    error(0, errno, "fork");
+    return false;
+  }
+  if(pid == 0){
+    /* Child */
+    if(daemonize){
+      if(not become_a_daemon()){
+	_exit(EX_OSERR);
+      }
+    }
+
+    char **new_argv = NULL;
+    char *tmp;
+    int i = 0;
+    for (; argv[i]!=(char *)NULL; i++){
+      tmp = realloc(new_argv, sizeof(const char *) * ((size_t)i + 1));
+      if (tmp == NULL){
+	error(0, errno, "realloc");
+	free(new_argv);
+	_exit(EXIT_FAILURE);
+      }
+      new_argv = (char **)tmp;
+      new_argv[i] = strdup(argv[i]);
+    }
+    new_argv[i] = (char *) NULL;
+    
+    execv(path, (char *const *)new_argv);
+    error(0, errno, "execv");
+    _exit(EXIT_FAILURE);
+  }
+  if(pid_return != NULL){
+    *pid_return = pid;
+  }
+  do {
+    ret = waitpid(pid, &status, 0);
+  } while(ret == -1 and errno == EINTR
+	  and ((not interrupted_by_signal)
+	       or (not interruptable)));
+  if(interrupted_by_signal and interruptable){
+    return false;
+  }
+  if(ret == -1){
+    error(0, errno, "waitpid");
+    return false;
+  }
+  if(WIFEXITED(status) and WEXITSTATUS(status) == 0){
+    return true;
+  }
+  return false;
+}
+
+int is_plymouth(const struct dirent *proc_entry){
+  int ret;
+  {
+    uintmax_t maxvalue;
+    char *tmp;
+    errno = 0;
+    maxvalue = strtoumax(proc_entry->d_name, &tmp, 10);
+
+    if(errno != 0 or *tmp != '\0' or maxvalue != (uintmax_t)((pid_t)maxvalue)){
+      return 0;
+    }
+  }
+  char exe_target[sizeof(plymouth_path)];
+  char *exe_link;
+  ret = asprintf(&exe_link, "/proc/%s/exe", proc_entry->d_name);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    return 0;
+  }
+
+  struct stat exe_stat;
+  ret = lstat(exe_link, &exe_stat);
+  if(ret == -1){
+    free(exe_link);
+    if(errno != ENOENT){
+      error(0, errno, "lstat");
+    }
+    return 0;
+  }
+  
+  if(not S_ISLNK(exe_stat.st_mode)
+     or exe_stat.st_uid != 0
+     or exe_stat.st_gid != 0){
+    free(exe_link);
+    return 0;
+  }
+
+  ssize_t sret = readlink(exe_link, exe_target, sizeof(exe_target));
+  free(exe_link);
+  if((sret != (ssize_t)sizeof(plymouth_path)-1) or
+      (memcmp(plymouth_path, exe_target,
+	      sizeof(plymouth_path)-1) != 0)){
+    return 0;
+  }
+  return 1;
+}
+
+pid_t get_pid(void){
+  int ret;
+  FILE *pidfile = fopen(plymouth_pid, "r");
+  uintmax_t maxvalue = 0;
+  if(pidfile != NULL){
+    ret = fscanf(pidfile, "%" SCNuMAX, &maxvalue);
+    if(ret != 1){
+      maxvalue = 0;
+    }
+    fclose(pidfile);
+  }
+  if(maxvalue == 0){
+    struct dirent **direntries;
+    ret = scandir("/proc", &direntries, is_plymouth, alphasort);
+    sscanf(direntries[0]->d_name, "%" SCNuMAX, &maxvalue);
+  }
+  pid_t pid;
+  pid = (pid_t)maxvalue;
+  if((uintmax_t)pid == maxvalue){
+    return pid;
+  }
+  
+  return 0;
+}
+
+const char **getargv(pid_t pid){
+  int cl_fd;
+  char *cmdline_filename;
+  ssize_t sret;
+  int ret;
+  
+  ret = asprintf(&cmdline_filename, "/proc/%" PRIuMAX "/cmdline",
+		 (uintmax_t)pid);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    return NULL;
+  }
+  
+  /* Open /proc/<pid>/cmdline  */
+  cl_fd = open(cmdline_filename, O_RDONLY);
+  free(cmdline_filename);
+  if(cl_fd == -1){
+    error(0, errno, "open");
+    return NULL;
+  }
+  
+  size_t cmdline_allocated = 0;
+  size_t cmdline_len = 0;
+  char *cmdline = NULL;
+  char *tmp;
+  const size_t blocksize = 1024;
+  do {
+    /* Allocate more space? */
+    if(cmdline_len + blocksize > cmdline_allocated){
+      tmp = realloc(cmdline, cmdline_allocated + blocksize);
+      if(tmp == NULL){
+	error(0, errno, "realloc");
+	free(cmdline);
+	close(cl_fd);
+	return NULL;
+      }
+      cmdline = tmp;
+      cmdline_allocated += blocksize;
+    }
+    
+    /* Read data */
+    sret = read(cl_fd, cmdline + cmdline_len,
+		cmdline_allocated - cmdline_len);
+    if(sret == -1){
+      error(0, errno, "read");
+      free(cmdline);
+      close(cl_fd);
+      return NULL;
+    }
+    cmdline_len += (size_t)sret;
+  } while(sret != 0);
+  ret = close(cl_fd);
+  if(ret == -1){
+    error(0, errno, "close");
+    free(cmdline);
+    return NULL;
+  }
+  
+  /* we got cmdline and cmdline_len, ignore rest... */
+  const char **argv = NULL;
+  size_t argv_size = 0;
+  for(char *arg = cmdline; arg-cmdline < (ssize_t)cmdline_len;
+      arg = strchr(arg, '\0')+1){
+    tmp = realloc(argv, ((++argv_size)+1)*sizeof(char *));
+    if(tmp == NULL){
+      error(0, errno, "realloc");
+      free(argv);
+      return NULL;
+    }
+    argv = (const char **)tmp;
+    argv[argv_size-1] = arg;
+  }
+  argv[argv_size] = NULL;
+  return argv;
+}
+
+int main(__attribute__((unused))int argc,
+	 __attribute__((unused))char **argv){
+  char *prompt;
+  char *prompt_arg;
+  pid_t plymouth_command_pid;
+  int ret;
+  bool bret;
+
+  /* test -x /bin/plymouth */
+  ret = access(plymouth_path, X_OK);
+  if(ret == -1){
+    exit(EXIT_FAILURE);
+  }
+
+  { /* Add signal handlers */
+    struct sigaction old_action,
+      new_action = { .sa_handler = termination_handler,
+		     .sa_flags = 0 };
+    sigemptyset(&new_action.sa_mask);
+    for(int *sig = (int[]){ SIGINT, SIGHUP, SIGTERM, 0 }; *sig != 0; sig++){
+      ret = sigaddset(&new_action.sa_mask, *sig);
+      if(ret == -1){
+	error(0, errno, "sigaddset");
+	exit(EX_OSERR);
+      }
+      ret = sigaction(*sig, NULL, &old_action);
+      if(ret == -1){
+	error(0, errno, "sigaction");
+	exit(EX_OSERR);
+      }
+      if(old_action.sa_handler != SIG_IGN){
+	ret = sigaction(*sig, &new_action, NULL);
+	if(ret == -1){
+	  error(0, errno, "sigaction");
+	  exit(EX_OSERR);
+	}
+      }
+    }
+  }
+    
+  /* plymouth --ping */
+  bret = exec_and_wait(&plymouth_command_pid, plymouth_path,
+		       (const char *[]){ (const char *)plymouth_path, (const char *)"--ping", (const char *)NULL},
+		       true, false);
+  if(not bret){
+    if(interrupted_by_signal){
+      kill_and_wait(plymouth_command_pid);
+    }
+    exit(EXIT_FAILURE);
+  }
+  
+  prompt = makeprompt();
+  ret = asprintf(&prompt_arg, "--prompt=%s", prompt);
+  free(prompt);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    exit(EXIT_FAILURE);
+  }
+  
+  /* plymouth ask-for-password --prompt="$prompt" */
+  bret = exec_and_wait(&plymouth_command_pid, plymouth_path,
+		       (const char *[]){plymouth_path, "ask-for-password", prompt_arg, NULL},
+		       true, false);
+  free(prompt_arg);
+  if(not bret){
+    if(interrupted_by_signal){
+      kill_and_wait(plymouth_command_pid);
+    } else {
+      exit(EXIT_FAILURE);
+    }
+  }
+  
+  if(bret){
+    exit(EXIT_SUCCESS);
+  }
+  
+  const char **plymouthd_argv = NULL;
+  pid_t pid = get_pid();
+  if(pid == 0){
+    error(0, 0, "plymouthd pid not found");
+  } else {
+    plymouthd_argv = getargv(pid);
+  }
+  if(plymouthd_argv == NULL){
+    plymouthd_argv = plymouthd_default_argv;
+  }
+  
+  bret = exec_and_wait(NULL, plymouth_path,
+  		       (const char *[]){plymouth_path, "quit", NULL}, false, false);
+  if(not bret){
+    exit(EXIT_FAILURE);
+  }
+  bret = exec_and_wait(NULL, plymouthd_path, plymouthd_argv, false, true);
+  if(not bret){
+    exit(EXIT_FAILURE);
+  }
+  exec_and_wait(NULL, plymouth_path,
+  		(const char *[]){ plymouth_path, "show-splash", NULL }, false, false);
+  exit(EXIT_FAILURE);
+}

=== modified file 'plugins.d/splashy.c'
--- plugins.d/splashy.c	2009-10-18 16:09:05 +0000
+++ plugins.d/splashy.c	2010-09-07 16:48:58 +0000
@@ -29,7 +29,7 @@
 				   SIG_IGN, kill(), SIGKILL */
 #include <stddef.h>		/* NULL */
 #include <stdlib.h>		/* getenv() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf() */
 #include <stdlib.h>		/* EXIT_FAILURE, free(),
 				   EXIT_SUCCESS */
 #include <sys/types.h>		/* pid_t, DIR, struct dirent,
@@ -49,6 +49,7 @@
 				   E2BIG, EFAULT, EIO, ETXTBSY,
 				   EISDIR, ELIBBAD, EPERM, EINTR,
 				   ECHILD */
+#include <error.h>		/* error() */
 #include <sys/wait.h>		/* waitpid(), WIFEXITED(),
 				   WEXITSTATUS() */
 #include <sysexits.h>		/* EX_OSERR, EX_OSFILE,
@@ -109,7 +110,7 @@
     proc_dir = opendir("/proc");
     if(proc_dir == NULL){
       int e = errno;
-      perror("opendir");
+      error(0, errno, "opendir");
       switch(e){
       case EACCES:
       case ENOTDIR:
@@ -151,7 +152,7 @@
 	char *exe_link;
 	ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  exitstatus = EX_OSERR;
 	  goto failure;
 	}
@@ -165,7 +166,7 @@
 	    continue;
 	  }
 	  int e = errno;
-	  perror("lstat");
+	  error(0, errno, "lstat");
 	  free(exe_link);
 	  switch(e){
 	  case EACCES:
@@ -213,60 +214,60 @@
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
@@ -283,7 +284,7 @@
     goto failure;
   }
   if(splashy_command_pid == -1){
-    perror("fork");
+    error(0, errno, "fork");
     exitstatus = EX_OSERR;
     goto failure;
   }
@@ -293,7 +294,7 @@
       const char splashy_command[] = "/sbin/splashy_update";
       execl(splashy_command, splashy_command, prompt, (char *)NULL);
       int e = errno;
-      perror("execl");
+      error(0, errno, "execl");
       switch(e){
       case EACCES:
       case ENOENT:
@@ -341,7 +342,7 @@
       goto failure;
     }
     if(ret == -1){
-      perror("waitpid");
+      error(0, errno, "waitpid");
       if(errno == ECHILD){
 	splashy_command_pid = 0;
       }
@@ -379,27 +380,27 @@
 	 the real user ID (_mandos) */
       ret = setuid(geteuid());
       if(ret == -1){
-	perror("setuid");
+	error(0, errno, "setuid");
       }
       
       setsid();
       ret = chdir("/");
       if(ret == -1){
-	perror("chdir");
+	error(0, errno, "chdir");
       }
 /*       if(fork() != 0){ */
 /* 	_exit(EXIT_SUCCESS); */
 /*       } */
       ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace stdout */
       if(ret == -1){
-	perror("dup2");
+	error(0, errno, "dup2");
 	_exit(EX_OSERR);
       }
       
       execl("/sbin/splashy", "/sbin/splashy", "boot", (char *)NULL);
       {
 	int e = errno;
-	perror("execl");
+	error(0, errno, "execl");
 	switch(e){
 	case EACCES:
 	case ENOENT:
@@ -425,13 +426,13 @@
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());

=== modified file 'plugins.d/usplash.c'
--- plugins.d/usplash.c	2010-05-18 16:56:54 +0000
+++ plugins.d/usplash.c	2010-09-07 16:48:58 +0000
@@ -31,19 +31,20 @@
 #include <fcntl.h>		/* open(), O_WRONLY, O_RDONLY */
 #include <iso646.h>		/* and, or, not*/
 #include <errno.h>		/* errno, EINTR */
+#include <error.h>
 #include <sys/types.h>		/* size_t, ssize_t, pid_t, DIR, struct
 				   dirent */
 #include <stddef.h>		/* NULL */
 #include <string.h>		/* strlen(), memcmp() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf()*/
 #include <unistd.h>		/* close(), write(), readlink(),
 				   read(), STDOUT_FILENO, sleep(),
 				   fork(), setuid(), geteuid(),
 				   setsid(), chdir(), dup2(),
 				   STDERR_FILENO, execv() */
 #include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
-				   EXIT_SUCCESS, malloc(), _exit() */
-#include <stdlib.h>		/* getenv() */
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
 #include <dirent.h>		/* opendir(), readdir(), closedir() */
 #include <inttypes.h>		/* intmax_t, strtoimax() */
 #include <sys/stat.h>		/* struct stat, lstat(), S_ISLNK */
@@ -153,7 +154,7 @@
   size_t cmdline_len = 0;
   DIR *proc_dir = opendir("/proc");
   if(proc_dir == NULL){
-    perror("opendir");
+    error(0, errno, "opendir");
     return -1;
   }
   errno = 0;
@@ -181,7 +182,7 @@
       char *exe_link;
       ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
       if(ret == -1){
-	perror("asprintf");
+	error(0, errno, "asprintf");
 	goto fail_find_usplash;
       }
       
@@ -193,7 +194,7 @@
 	  free(exe_link);
 	  continue;
 	}
-	perror("lstat");
+	error(0, errno, "lstat");
 	free(exe_link);
 	goto fail_find_usplash;
       }
@@ -224,13 +225,13 @@
 	ret = asprintf(&cmdline_filename, "/proc/%s/cmdline",
 		       proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  goto fail_find_usplash;
 	}
 	cl_fd = open(cmdline_filename, O_RDONLY);
 	free(cmdline_filename);
 	if(cl_fd == -1){
-	  perror("open");
+	  error(0, errno, "open");
 	  goto fail_find_usplash;
 	}
       }
@@ -242,7 +243,7 @@
 	if(cmdline_len + blocksize > cmdline_allocated){
 	  tmp = realloc(cmdline, cmdline_allocated + blocksize);
 	  if(tmp == NULL){
-	    perror("realloc");
+	    error(0, errno, "realloc");
 	    close(cl_fd);
 	    goto fail_find_usplash;
 	  }
@@ -253,7 +254,7 @@
 	sret = read(cl_fd, cmdline + cmdline_len,
 		    cmdline_allocated - cmdline_len);
 	if(sret == -1){
-	  perror("read");
+	  error(0, errno, "read");
 	  close(cl_fd);
 	  goto fail_find_usplash;
 	}
@@ -261,14 +262,14 @@
       } while(sret != 0);
       ret = close(cl_fd);
       if(ret == -1){
-	perror("close");
+	error(0, errno, "close");
 	goto fail_find_usplash;
       }
     }
     /* Close directory */
     ret = closedir(proc_dir);
     if(ret == -1){
-      perror("closedir");
+      error(0, errno, "closedir");
       goto fail_find_usplash;
     }
     /* Success */
@@ -323,26 +324,26 @@
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -351,7 +352,7 @@
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -360,7 +361,7 @@
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -369,7 +370,7 @@
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -378,7 +379,7 @@
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -387,7 +388,7 @@
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -399,7 +400,7 @@
   /* Write command to FIFO */
   if(not usplash_write(&fifo_fd, "TIMEOUT", "0")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -411,7 +412,7 @@
   
   if(not usplash_write(&fifo_fd, "INPUTQUIET", prompt)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -429,7 +430,7 @@
   outfifo_fd = open("/dev/.initramfs/usplash_outfifo", O_RDONLY);
   if(outfifo_fd == -1){
     if(errno != EINTR){
-      perror("open");
+      error(0, errno, "open");
       status = EX_OSERR;
     }
     goto failure;
@@ -448,7 +449,7 @@
       char *tmp = realloc(buf, buf_allocated + blocksize);
       if(tmp == NULL){
 	if(errno != EINTR){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -460,7 +461,7 @@
 		buf_allocated - buf_len);
     if(sret == -1){
       if(errno != EINTR){
-	perror("read");
+	error(0, errno, "read");
 	status = EX_OSERR;
       }
       TEMP_FAILURE_RETRY(close(outfifo_fd));
@@ -475,7 +476,7 @@
   ret = close(outfifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -488,7 +489,7 @@
   
   if(not usplash_write(&fifo_fd, "TIMEOUT", "15")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -501,7 +502,7 @@
   ret = close(fifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -515,7 +516,7 @@
       sret = write(STDOUT_FILENO, buf + written, buf_len - written);
       if(sret == -1){
 	if(errno != EINTR){
-	  perror("write");
+	  error(0, errno, "write");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -552,7 +553,7 @@
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -561,7 +562,7 @@
   if(outfifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(outfifo_fd));
     if(ret == -1){
-      perror("close");
+      error(0, errno, "close");
     }
   }
   
@@ -575,7 +576,7 @@
 			   (sizeof(char *)
 			    * (size_t)(cmdline_argc + 2)));
       if(tmp == NULL){
-	perror("realloc");
+	error(0, errno, "realloc");
 	free(cmdline_argv);
 	return status;
       }
@@ -602,13 +603,13 @@
        the real user ID (_mandos) */
     ret = setuid(geteuid());
     if(ret == -1){
-      perror("setuid");
+      error(0, errno, "setuid");
     }
     
     setsid();
     ret = chdir("/");
     if(ret == -1){
-      perror("chdir");
+      error(0, errno, "chdir");
       _exit(EX_OSERR);
     }
 /*     if(fork() != 0){ */
@@ -616,13 +617,13 @@
 /*     } */
     ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
     if(ret == -1){
-      perror("dup2");
+      error(0, errno, "dup2");
       _exit(EX_OSERR);
     }
     
     execv(usplash_name, cmdline_argv);
     if(not interrupted_by_signal){
-      perror("execv");
+      error(0, errno, "execv");
     }
     free(cmdline);
     free(cmdline_argv);
@@ -633,7 +634,7 @@
   sleep(2);
   if(not usplash_write(&fifo_fd, "PULSATE", NULL)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
     }
   }
   
@@ -641,7 +642,7 @@
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -652,13 +653,13 @@
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());

